﻿using System;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Drawing.Text;
using System.Globalization;
using System.Text.RegularExpressions;
using System.Windows.Forms;

namespace Thumbnailstitcher
{
	public partial class Form1 : Form
	{
		public Form1()
		{
			InitializeComponent();
		}

		private void button1_Click_1(object sender, EventArgs e)
		{
			moviepathFileDialog.ShowDialog();
			filename.Lines = moviepathFileDialog.FileNames;
		}

		private void browse_thumb_Click(object sender, EventArgs e)
		{
			resultFileDialog.ShowDialog();
			result.Text = resultFileDialog.FileName;
		}
		private void drawString(Graphics g, string text, Color outlineColor, int OutlineWidth, Font font, Color color, Rectangle rect, StringFormat format)
		{
			// http://stackoverflow.com/questions/19842722/setting-a-font-with-outline-color-in-c-sharp
			using (GraphicsPath gp = new GraphicsPath())
			using (Pen outline = new Pen(outlineColor, OutlineWidth)
			{ LineJoin = LineJoin.Round })
			using (Brush foreBrush = new SolidBrush(color))
			{
				gp.AddString(text, font.FontFamily, (int)font.Style,
					font.Size, rect, format);
				//g.ScaleTransform(1.3f, 1.35f);
				g.SmoothingMode = SmoothingMode.HighQuality;
				g.DrawPath(outline, gp);
				g.FillPath(foreBrush, gp);
				g.ResetTransform();
			}
		}
		private void generate_Click(object sender, EventArgs e)
		{
			backgroundWorker1.RunWorkerAsync();
			generate.Enabled = false;
		}
		bool loading = true;
		private void captionFont_Click(object sender, EventArgs e)
		{
			if (!loading)
				captionFontDialog.ShowDialog();
			captionFont.Text = captionFontDialog.Font.Name + ", " + captionFontDialog.Font.Size + " pt";
		}

		private void backgroundColor_Click(object sender, EventArgs e)
		{
			if (!loading)
				colorDialogBackgroundColor.ShowDialog();
			backgroundColor.BackColor = colorDialogBackgroundColor.Color;
		}

		private void tsForeColor_Click(object sender, EventArgs e)
		{
			if (!loading)
				colorDialogtsForeColor.ShowDialog();
			tsForeColor.BackColor = colorDialogtsForeColor.Color;
		}

		private void tsOutlineColor_Click(object sender, EventArgs e)
		{
			if (!loading)
				colorDialogTsOutlineColor.ShowDialog();
			tsOutlineColor.BackColor = colorDialogTsOutlineColor.Color;
		}

		private void tsFont_Click(object sender, EventArgs e)
		{
			if (!loading)
				tsFontDialog.ShowDialog();
			tsFont.Text = tsFontDialog.Font.Name + ", " + tsFontDialog.Font.Size + " pt";
		}

		private void Form1_Load(object sender, EventArgs e)
		{
			var s = Properties.Settings.Default;
			pixelwidth.Value = s.width;
			colorDialogBackgroundColor.Color = s.bg_color;
			X.Value = s.x;
			Y.Value = s.y;
			checkBoxTS.Checked = s.check_ts;
			tsFontDialog.Font = s.ts_font;
			colorDialogtsForeColor.Color = s.ts_color;
			colorDialogTsOutlineColor.Color = s.ts_outline_color;
			alpha.Value = s.ts_alpha;
			strokeSize.Value = s.ts_outline_size;
			checkBoxCaption.Checked = s.check_cap;
			captionFontDialog.Font = s.cap_font;
			captionText.Text = s.cap_text;
			captionHeight.Value = s.cap_height;
			checkBoxKeyframe.Checked = s.check_I;
			paddingSize.Value = s.padding;
			captionFont_Click(this, null);
			tsFont_Click(this, null);
			backgroundColor_Click(this, null);
			tsForeColor_Click(this, null);
			tsOutlineColor_Click(this, null);
			loading = false;
		}

		private void Form1_FormClosing(object sender, FormClosingEventArgs e)
		{
			var s = Properties.Settings.Default;
			s.width = pixelwidth.Value;
			s.bg_color = colorDialogBackgroundColor.Color;
			s.x = X.Value;
			s.y = Y.Value;
			s.check_ts = checkBoxTS.Checked;
			s.ts_font = tsFontDialog.Font;
			s.ts_color = colorDialogtsForeColor.Color;
			s.ts_outline_color = colorDialogTsOutlineColor.Color;
			s.ts_alpha = alpha.Value;
			s.ts_outline_size = strokeSize.Value;
			s.check_cap = checkBoxCaption.Checked;
			s.cap_font = captionFontDialog.Font;
			s.cap_text = captionText.Text;
			s.cap_height = captionHeight.Value;
			s.check_I = checkBoxKeyframe.Checked;
			s.padding = paddingSize.Value;
			s.Save();
		}
		string shortenPrefix(double input)
		{
			if (input < 1e-30)
				return "";
			string[] prefix = { "y", "z", "a", "f", "n", "µ", "m", "", "k", "M", "G", "T", "P", "E", "Z", "Y" };
			int shift = 7;//By how many elements I have to shift to get the right prefix index
			int idx = (int)Math.Floor(Math.Log(input, 1024)) + shift;
			idx = Math.Min(prefix.Length - 1, Math.Max(0, idx));
			return prefix[idx];
		}
		double shortenMantissa(double input)
		{
			if (input < 1e-30)
				return 0;
			int idx = (int)Math.Floor(Math.Log(input, 1024));
			return input / Math.Pow(1024, idx);
		}
		private void backgroundWorker1_DoWork(object sender, System.ComponentModel.DoWorkEventArgs e)
		{
			// http://superuser.com/questions/538112/meaningful-thumbnails-for-a-video-using-ffmpeg
			/*
            for X in 1..N
                T = integer( (X - 0.5) * D / N )  
                run `ffmpeg -ss <T> -i <movie>
                    -vf select="eq(pict_type\,I)" -vframes 1 image<X>.jpg`
            */
			foreach (var file in filename.Lines)
			{
				Process proc = new Process();
				proc.StartInfo.RedirectStandardError = true;
				proc.StartInfo.FileName = "ffmpeg";
				proc.StartInfo.Arguments = "-y -i \"" + file + "\"";
				proc.StartInfo.UseShellExecute = false;
				proc.StartInfo.CreateNoWindow = true;
				proc.Start();
				proc.WaitForExit();
				string s = proc.StandardError.ReadToEnd();
				Match dur = new Regex(@"Duration: (\d+):(\d{2}):(\d{2}).(\d+)").Match(s);
				if (!dur.Success)
				{
					throw new Exception("Error determining duration");
				}
				Match ar = new Regex(@"DAR (\d+):(\d+)").Match(s);
				Match ar_2 = new Regex(@" (\d+)x(\d+)[ ,]").Match(s);
				double bitrate = Double.NaN;
				Double.TryParse(new Regex(@"bitrate: (\d+) kb/s").Match(s).Groups[1].Value, NumberStyles.Float, CultureInfo.InvariantCulture, out bitrate);
				bitrate *= 1024;

				double fps = Double.NaN;
				Double.TryParse(new Regex(@" (\d+(?:\.\d+)?) fps").Match(s).Groups[1].Value, NumberStyles.Float, CultureInfo.InvariantCulture, out fps);

				var b_v = Double.NaN;
				var b_a = Double.NaN;
				Double.TryParse(new Regex(@"^.*Video:.*?(\d+(?:\.\d+)?) kb/s.*$", RegexOptions.Multiline).Match(s).Groups[1].Value, NumberStyles.Float, CultureInfo.InvariantCulture, out b_v);
				Double.TryParse(new Regex(@"^.*Audio:.*?(\d+(?:\.\d+)?) kb/s.*$", RegexOptions.Multiline).Match(s).Groups[1].Value, NumberStyles.Float, CultureInfo.InvariantCulture, out b_a);
				b_v *= 1024;
				b_a *= 1024;

				var c_v = new Regex(@"Video: (\w+)").Match(s).Groups[1].Value.ToUpper();
				var c_a = new Regex(@"Audio: (\w+)").Match(s).Groups[1].Value.ToUpper();
				double aspect_ratio = 1;
				if (ar.Success)
					aspect_ratio = Double.Parse(ar.Groups[1].Value) / Double.Parse(ar.Groups[2].Value);
				else if (ar_2.Success)
					aspect_ratio = Double.Parse(ar_2.Groups[1].Value) / Double.Parse(ar_2.Groups[2].Value);

				int x = decimal.ToInt32(X.Value);
				int y = decimal.ToInt32(Y.Value);
				int n = x * y;

				int picwidth = decimal.ToInt32(pixelwidth.Value);
				int padding = decimal.ToInt32(paddingSize.Value);
				int picheight = (int)((picwidth - (x - 1) * padding) / aspect_ratio / x * y + (y - 1) * padding);
				int thumbwidth = (picwidth - (x - 1) * padding) / x;
				int thumbheight = (picheight - (y - 1) * padding) / y;
				int captionheight = checkBoxCaption.Checked ? decimal.ToInt32(captionHeight.Value) : 0;

				long durins = Int32.Parse(dur.Groups[1].Value) * 3600 + Int32.Parse(dur.Groups[2].Value) * 60 + Int32.Parse(dur.Groups[3].Value);

				Bitmap data = new Bitmap(picwidth, picheight + captionheight);
				Graphics g = Graphics.FromImage(data);
				g.Clear(Color.FromArgb(colorDialogBackgroundColor.Color.ToArgb() | (0xFF << 24)));

				g.SmoothingMode = SmoothingMode.AntiAlias;
				g.InterpolationMode = InterpolationMode.HighQualityBicubic;
				g.PixelOffsetMode = PixelOffsetMode.HighQuality;
				g.TextRenderingHint = TextRenderingHint.AntiAliasGridFit;

				if (checkBoxCaption.Checked)
					drawString(g, String.Format(
							CultureInfo.InvariantCulture,
							captionText.Text,// Format string

							file.Substring(file.LastIndexOf("\\") + 1),
							shortenMantissa(new System.IO.FileInfo(file).Length), shortenPrefix(new System.IO.FileInfo(file).Length), new System.IO.FileInfo(file).Length,
							ar_2.Groups[1].Value + "x" + ar_2.Groups[2].Value,
							dur.Groups[1].Value + ":" + dur.Groups[2].Value + ":" + dur.Groups[3].Value,
							shortenMantissa(bitrate), shortenPrefix(bitrate),
							fps,
							c_v.ToUpper(),
							c_a.ToUpper(),
							shortenMantissa(b_v), shortenPrefix(b_v),
							shortenMantissa(b_a), shortenPrefix(b_a)
						), colorDialogTsOutlineColor.Color,
						decimal.ToInt32(strokeSize.Value),
						captionFontDialog.Font, colorDialogtsForeColor.Color, new Rectangle(0, 0, picwidth, captionheight), new StringFormat());

				for (int j = 0; j < x; ++j)
				{
					for (int i = 0; i < y; ++i)
					{
						//Generate Thumbnail with ffmpeg
						Process ffmpeg = new Process();
						ffmpeg.StartInfo.RedirectStandardOutput = true;
						ffmpeg.StartInfo.UseShellExecute = false;
						ffmpeg.StartInfo.CreateNoWindow = true;
						ffmpeg.StartInfo.FileName = "ffmpeg";
						int ts = (int)((j * x + i + 0.5) * durins / n);
						ffmpeg.StartInfo.Arguments = "-loglevel warning -ss " + ts + " -i \"" + file + "\" " +
							(checkBoxKeyframe.Checked ? "-vf select=\"eq(pict_type\\,I)\" " : "") +
							"-vframes 1 -f rawvideo -vcodec rawvideo -pix_fmt bgr24 -s " +
							thumbwidth + "x" + thumbheight +
							" -";
						ffmpeg.Start();

						//read from pipe (stdout)
						byte[] pic = new byte[thumbwidth * thumbheight * 3];
						int read = 0;
						while (!ffmpeg.HasExited && read < thumbwidth * thumbheight * 3)
						{
							int val = ffmpeg.StandardOutput.BaseStream.Read(pic, read, thumbwidth * thumbheight * 3 - read);
							read += val;
						}

						//Draw it on bitmap
						var rect = new Rectangle(i * thumbwidth + padding * i, j * thumbheight + padding * j + captionheight, thumbwidth, thumbheight);
						BitmapData bd = data.LockBits(
							rect, ImageLockMode.WriteOnly, PixelFormat.Format24bppRgb);
						for (int k = 0; k < thumbheight; ++k)
							System.Runtime.InteropServices.Marshal.Copy(pic, k * thumbwidth * 3, bd.Scan0 + k * bd.Stride, thumbwidth * 3);
						data.UnlockBits(bd);
						if (checkBoxTS.Checked)
						{
							//overlay timestamp
							var format = new StringFormat();
							format.Alignment = StringAlignment.Center;
							format.LineAlignment = StringAlignment.Far;
							drawString(g, String.Format("{0:00}:{1:00}:{2:00}", ts / 3600, (ts / 60) % 60, ts % 60),
								Color.FromArgb(decimal.ToInt32(alpha.Value), colorDialogTsOutlineColor.Color.R, colorDialogTsOutlineColor.Color.G, colorDialogTsOutlineColor.Color.B),
								decimal.ToInt32(strokeSize.Value),
								tsFontDialog.Font,
								Color.FromArgb(decimal.ToInt32(alpha.Value), colorDialogtsForeColor.Color.R, colorDialogtsForeColor.Color.G, colorDialogtsForeColor.Color.B),
								rect, format);
						}
						backgroundWorker1.ReportProgress(100 * (j * x + i) / n);
					}
				}
				var resultpath = "";
				if (filename.Lines.Length > 1)
				{
					resultpath = result.Text.Substring(0, result.Text.LastIndexOf("\\") + 1) +
						file.Substring(file.LastIndexOf("\\")) + result.Text.Substring(result.Text.LastIndexOf("."));
				}
				else
					resultpath = result.Text;
				data.Save(resultpath);
				data.Dispose();//Necessary for a sustainable memory footprint
			}
			backgroundWorker1.ReportProgress(0);
		}

		private void backgroundWorker1_ProgressChanged(object sender, System.ComponentModel.ProgressChangedEventArgs e)
		{
			progressBar1.Value = e.ProgressPercentage;
		}

		private void backgroundWorker1_RunWorkerCompleted(object sender, System.ComponentModel.RunWorkerCompletedEventArgs e)
		{
			generate.Enabled = true;
		}
	}
}
